# Program for a basic GA
# This simple GA is based on the algorithm proposed by Coley, David A.
# An Introduction to Genetic Algorithms for Scientists and Engineers.
# Singapore: World Scientific Publishing Co. Pte. Ltd., 1999. 227 pp.
# See page 12 of the above reference.
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# Porto Alegre, November, 12th 2016.
#
def get_max(x):
    """Function to return the maximum value for an array"""
    
    # Import library
    import numpy as np
    
    # Find maximum
    max1 = np.max(x)
    
    return max1

def get_mean(x):
    """Function to return the mean value for an array"""
    
    # Import library
    import numpy as np
    
    # Find mean value
    mean1 = np.mean(x)
    
    return mean1
    
def gen_bin_strings(pop_size,len_str):
    """Function to generate binary strings"""
    
    # Import library
    import numpy as np
    
    # Set up empty list
    bin_str_list = []
    
    # Generate binary string
    for i in range(pop_size):
        bin1 = np.random.randint(2,size=len_str)    # Binary (2) strings of length of len_str
        bin_str0 = str(bin1)
        bin_str0 = bin_str0.replace(" ","")
        bin_str0 = bin_str0.replace("[","")
        bin_str0 = bin_str0.replace("]","")
        bin_str_list.append(bin_str0)
        
    return bin_str_list

def bin2dec(string_num):
    """Function to convert binary string to decimal integer"""
    
    # Convert binary to decinal
    num1 = str(int(string_num,2))
    
    return num1

def bin2dec_array(pop_in):
    """Function to convert a list of binary strings to an array of integer decimals"""
    
    # Import libray
    import numpy as np
    
    # Set up an array of zeros with length of pop_in
    x = np.zeros(len(pop_in))
    
    # Looping through population to convert from binary to an array of decimals 
    i = 0
    for line in pop_in:
        aux = bin2dec(line)
        x[i] = aux
        i+=1
    
    return x
    
def fitness_func(x):
    """Function to calculate fitness function"""
    
    # Fitness function
    y = x*x
    
    return y

def reverse_array(x):
    """Function to reverse array"""
    
    # Reverse array
    x_rev = x[::-1]
    
    return x_rev

def sort_popupation(f1,f0,pop_in):
    """Function to sort population"""
    
    # Set up empty lists
    in_list = []
    pop_out = []
    
    # Convert array to list
    my_list =  list(f0)
    
    # Looping through f1
    for line in f1:
        ind1 = my_list.index(line)
        in_list.append(ind1)
    
    # Looping through list of indeces
    for line in in_list:
        index1 = int(line)
        pop_out.append(pop_in[index1])
     
    return pop_out
        
def shuffle(pop_in):
    """Function to shuffle population"""
    
    # Import library
    import random
    
    # Shuffle by calling random.shuffle()
    random.shuffle(pop_in)
            
    return pop_in

def generate_offsprings(str1,str2):
    """Function to generate offsprings"""
    
    # Import library
    import random
    
    # Define partition locus
    locus = random.randint(2,len(str1)-1)
    
    # Generate offspring
    offspring1 = str1[0:locus] + str2[locus:]
    offspring2 = str2[0:locus] + str1[locus:]
    
    return offspring1, offspring2
    
def selection(pop_in):
    """Function for selection operator"""
    
    # Import library
    import numpy as np
            
    # Call bin2dec_array()
    x = bin2dec_array(pop_in)
    
    # Call fitness_func()
    fit = fitness_func(x)
    
    # Sort array
    fit_sorted = np.sort(fit)
    
    # Reverse fitness array
    fit_reverse = reverse_array(fit_sorted)
    
    # Sort population
    pop_in = sort_popupation(fit_reverse,fit,pop_in)
    
    return pop_in
    
def cross_over(pop_in,prob_in,div_in,size_in1):
    """Function to apply cross-over operator to a population. div_in indicate the fraction that will be
    used for cross_over"""
    
    # Import library
    import random 
    
    # Set up empty list
    new_pop = []
            
    # Looping throguh population
    for i in range( int( size_in1/div_in)  ):
        new_pop.append(pop_in[i])
    
    # Call shuffle()
    pop_out = shuffle(new_pop)

    # Get population size after shuffle
    size_in2 = len(pop_out)
    count_pop = size_in2
    
    # Get random number
    r_n = random.random()
    
    # Loop while to generate new population with offspring
    while count_pop < size_in1:         # To make sure the size of population stay the same
        if float(prob_in) < r_n:        # Test cross over prob            
            # Call generate_offsprings()
            for i in range(int(size_in2/2)):
                o1,o2 = generate_offsprings(pop_out[i],pop_out[i+1])
                pop_out.append(o1)
                pop_out.append(o2)
                count_pop += 2
        r_n = random.random()
    
    return pop_out
        
def mutation(pop_in,p_mut,pop_size,len_str):
    """Function to apply mutation operator"""
    
    # Import library
    import random
    
    # Determine total
    total = pop_size*len_str
    
    # Some editing list-> string
    str_pop = "".join(pop_in)
    str_pop = str_pop.replace("\n","")
    
    # Set up empty string
    str_pop_out = ""
    
    # Carry out mutation
    for i in range(total):
        if i == int(1/p_mut):
            if str_pop[i] == 1:
                str_pop_out+="0"
            else:
                str_pop_out+="1"
        else:
            str_pop_out+=str_pop[i]
    
    # Back to string
    # Set up empty list
    pop = []
    
    # Loop to recover population after mutation
    for i in range(pop_size):
        aux = ""
        for j in range(len_str):
            aux += str_pop_out[i+j]
        pop.append(aux)
    
    return pop
    
def plot_ga_evolution(y1,y2,n_iter,x_label,y_label,legend1,legend2):
    """Function to generate a plot of the evolution of the fitness function values for each iteraction"""
    
    # Import libraries
    import numpy as np
    import matplotlib.pyplot as plt
    
    # Generate array for x-axis
    x = np.linspace(1,n_iter)
    
    # Generate plot
    plt.plot(x,y1,label=legend1)
    plt.plot(x,y2,label=legend2)
    plt.legend(loc = "lower right")
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.grid(True)
    plt.show()
    plt.savefig("ga01.png")
    
def basic_ga(pop_in,pop_size,max_iter,p_cross,len_str,p_mut):  
    
    # Import library
    import numpy as np
    
    # Set up arrays of zeros
    max_array = np.zeros(max_iter)
    mean_array = np.zeros(max_iter)
    
    # Set up the fraction of the population
    div_in = 2 # 2 for half, 3 for one third
    
    # Loop for the number of iteractions
    for i in range(max_iter):
    
        # Call selection()
        pop_in = selection(pop_in)
        
        # Call cross_over()
        pop_in = cross_over(pop_in,p_cross,div_in,pop_size)
    
        # Call mutation()
        pop_in = mutation(pop_in,p_mut,pop_size,len_str)
        
        # Call bin2dec_array()
        d = bin2dec_array(pop_in)
        
        # Call fitness_func()
        fit = fitness_func(d)
        
        # Call get_max()
        max_array[i] = get_max(fit)
        
        # Call get_mean()
        mean_array[i] = get_mean(fit)
    
    # Call plot_ga_evolution()
    plot_ga_evolution(max_array,mean_array,max_iter,"Generation","Fitness function, f",
    "Maximum value of f","Mean value of f")
    
    return pop_in

def main():  
    
    # Set up initial values for GA
    pop_size    = 8            # Population size
    max_iter    = 50           # Number of iterations
    p_cross     = 0.8          # Probability of cross-over
    len_str     = 12           # Length of strings
    p_mut       = 0.17         # Probability of mutation
    
    # Call gen_bin_strings()
    pop_in = gen_bin_strings(pop_size,len_str)
        
    # Call basic_ga()
    pop = basic_ga(pop_in,pop_size,max_iter,p_cross,len_str,p_mut)
    
    # Show final population
    print("\nFinal population")
    for line in pop:
        print(line)
    
main()