# Program to calculate Principal component analysis coefficients
#
# Equations taken from:
# http://biol09.biol.umontreal.ca/PLcourses/PCA_algebra.pdf  (data03.csv)
#
def corr_matrix(A):
    """Function to determine the correlation matrix"""
    
    # Import libraries
    import numpy as np
    from scipy.stats import pearsonr
    
    # Get the number of rows and columns
    _,columns = np.shape(A)
    
    # Set up a matrix
    B = np.array( [[0]*columns]*columns,float)
    
    # Looping rows and columns of matrix A
    print("\nCorrelation matrix:")
    for i in range(columns):
        for j in range(columns):
            corr,_ = pearsonr(A[:,i],A[:,j])
            B[i][j] = corr    
            print(B[i][j],end=" ")
        print()
        
    return B

def eigen_val_vect(A):
    """Function to calculate eigenvalues and eigenvectors for a given nxn matrix"""
    
    # Import library
    import numpy.linalg as linalg
    
    # Get eigenvectors 
    eVal, eVec = linalg.eig(A)
    
    # Show results
    print("\nEigenvalues: ",eVal)
    print("\nMatrix of eigenvectors:")
    print(eVec)
    
    return eVal, eVec

def read_CSV_2_Matrix(file_in):
    """Function to read a CSV file and return a matrix"""
    
    # Import libraries
    import sys
    import scipy as sp
    import numpy as np
    
    # Try to open CSV file
    try:
        data = sp.genfromtxt(file_in, delimiter=",")
    except IOError:
        sys.exit("I can't find "+file_in+" file!")
 
    x = data[1:,1:]  # data[1:,1:] for second column and second line
    
    # Get number of rows and columns
    rows= len(x[:])
    columns = len(x[0])
    
    # Define a zero matrix
    A = np.array([[0]*(columns)]*rows,float)  
    
    # Looping through data to get a matrix
    print("\nData:")
    for i in range(rows):
        aux = x[i]
        for j in range(columns):
            A[i][j] = aux[j]
            print(A[i][j],end=" ")
        print()
    return A

def zero_mean_data(X):
    """Function to subtract the mean of column from each element in a column"""
    
    # Import library
    import numpy as np
    
    # Get rows and columns
    rows, columns = np.shape(X)
    
    # Set up a zero array
    my_mean = np.zeros(columns,float)
    
    # Define a matrix 
    Y = np.array([[0]*columns]*rows,float )
    
    # Looping through columns to get the mean for each column
    for i in range(columns):
        my_mean[i] =  np.mean(X[:,i])
    
    # Looping through data to get zero mean data
    print("\nZero-mean data:")
    for i in range(rows):
        for j in range(columns):
            Y[i][j] = X[i][j] - my_mean[j]
            print(Y[i][j],end=" ")
        print()
    
    return Y

def covariance_matrix(Y):
    """Function to calculate covariance matrix"""
    
    # Import library
    import numpy as np
    
    # Get rows and columns
    rows, columns = np.shape(Y)
    
    # Determine the transpose
    YT = np.transpose(Y)
    
    # Determine the covariance matrix (S)
    S = (1/(rows-1))*np.dot(YT,Y)
    
    # Show results
    print("\nCovariance matrix:")
    for i in range(columns):
        for j in range(columns):
            print(S[i][j],end=" ")
        print()
    
    return S 
     
def PCA(YC,U):
    """Function to calculate principal components"""
    
    # Import library
    import numpy as np
    
    # Principal components
    F = np.dot(YC,U)
    
    # Show results
    print("\nPrincipal components:")
    print(F)

def main():
    # Data file name
    data_file = "aa_table01.csv"
    
    # Call read_csv()
    X = read_CSV_2_Matrix(data_file)
    
    # Call corr_matrix()
    corr_matrix(X)
    
    # Call zero_mean_data()
    YC = zero_mean_data(X)
    
    # Call covariance_matrix()
    S = covariance_matrix(YC)
    
    # Call eigen_val_vect()
    _,U = eigen_val_vect(S)
        
    # Call PCA()
    PCA(YC,U)
    
main()
